@@ -3,6 +3,7 @@ package archive
 import (
 	"archive/tar"
 	"bytes"
+	"errors"
 	"fmt"
 	"io"
 	"net"
@@ -12,21 +13,13 @@ import (
 	"runtime"
 	"strings"
 	"testing"
+	"time"
 
 	"github.com/containers/storage/pkg/idtools"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
 )
 
-var tmp string
-
-func init() {
-	tmp = "/tmp/"
-	if runtime.GOOS == windows {
-		tmp = os.Getenv("TEMP") + `\`
-	}
-}
-
 var defaultArchiver = NewDefaultArchiver()
 
 func defaultTarUntar(src, dst string) error {
@@ -46,7 +39,9 @@ func defaultCopyWithTar(src, dst string) error {
 }
 
 func TestIsArchivePathDir(t *testing.T) {
-	cmd := exec.Command("sh", "-c", "mkdir -p /tmp/archivedir")
+	tmp := t.TempDir()
+	cmd := exec.Command("sh", "-c", "mkdir -p archivedir")
+	cmd.Dir = tmp
 	output, err := cmd.CombinedOutput()
 	if err != nil {
 		t.Fatalf("Fail to create an archive file for test : %s.", output)
@@ -57,15 +52,17 @@ func TestIsArchivePathDir(t *testing.T) {
 }
 
 func TestIsArchivePathInvalidFile(t *testing.T) {
-	cmd := exec.Command("sh", "-c", "dd if=/dev/zero bs=1024 count=1 of=/tmp/archive && gzip --stdout /tmp/archive > /tmp/archive.gz")
+	tmp := t.TempDir()
+	cmd := exec.Command("sh", "-c", "dd if=/dev/zero bs=1024 count=1 of=archive && gzip --stdout archive > archive.gz")
+	cmd.Dir = tmp
 	output, err := cmd.CombinedOutput()
 	if err != nil {
 		t.Fatalf("Fail to create an archive file for test : %s.", output)
 	}
-	if IsArchivePath(tmp + "archive") {
+	if IsArchivePath(filepath.Join(tmp, "archive")) {
 		t.Fatalf("Incorrectly recognised invalid tar path as archive")
 	}
-	if IsArchivePath(tmp + "archive.gz") {
+	if IsArchivePath(filepath.Join(tmp, "archive.gz")) {
 		t.Fatalf("Incorrectly recognised invalid compressed tar path as archive")
 	}
 }
@@ -77,29 +74,33 @@ func TestIsArchivePathTar(t *testing.T) {
 	} else {
 		whichTar = "tar"
 	}
-	cmdStr := fmt.Sprintf("touch /tmp/archivedata && %s -cf /tmp/archive /tmp/archivedata && gzip --stdout /tmp/archive > /tmp/archive.gz", whichTar)
+	cmdStr := fmt.Sprintf("touch archivedata && %s -cf archive archivedata && gzip --stdout archive > archive.gz", whichTar)
 	cmd := exec.Command("sh", "-c", cmdStr)
+	tmp := t.TempDir()
+	cmd.Dir = tmp
 	output, err := cmd.CombinedOutput()
 	if err != nil {
 		t.Fatalf("Fail to create an archive file for test : %s.", output)
 	}
-	if !IsArchivePath(tmp + "/archive") {
+	if !IsArchivePath(filepath.Join(tmp, "archive")) {
 		t.Fatalf("Did not recognise valid tar path as archive")
 	}
-	if !IsArchivePath(tmp + "archive.gz") {
+	if !IsArchivePath(filepath.Join(tmp, "archive.gz")) {
 		t.Fatalf("Did not recognise valid compressed tar path as archive")
 	}
 }
 
 func testDecompressStream(t *testing.T, ext, compressCommand string) {
+	tmp := t.TempDir()
 	cmd := exec.Command("sh", "-c",
-		fmt.Sprintf("touch /tmp/archive && %s /tmp/archive", compressCommand))
+		fmt.Sprintf("touch archive && %s archive", compressCommand))
+	cmd.Dir = tmp
 	output, err := cmd.CombinedOutput()
 	if err != nil {
 		t.Fatalf("Failed to create an archive file for test : %s.", output)
 	}
-	filename := "archive." + ext
-	archive, err := os.Open(tmp + filename)
+	filename := filepath.Join(tmp, "archive."+ext)
+	archive, err := os.Open(filename)
 	if err != nil {
 		t.Fatalf("Failed to open file %s: %v", filename, err)
 	}
@@ -133,41 +134,20 @@ func TestDecompressStreamXz(t *testing.T) {
 }
 
 func TestCompressStreamXzUnsupported(t *testing.T) {
-	dest, err := os.Create(tmp + "dest")
-	if err != nil {
-		t.Fatalf("Fail to create the destination file: %v", err)
-	}
-	defer dest.Close()
-
-	_, err = CompressStream(dest, Xz)
-	if err == nil {
+	if _, err := CompressStream(&bytes.Buffer{}, Xz); err == nil {
 		t.Fatalf("Should fail as xz is unsupported for compression format.")
 	}
 }
 
 func TestCompressStreamBzip2Unsupported(t *testing.T) {
-	dest, err := os.Create(tmp + "dest")
-	if err != nil {
-		t.Fatalf("Fail to create the destination file: %v", err)
-	}
-	defer dest.Close()
-
-	_, err = CompressStream(dest, Xz)
-	if err == nil {
-		t.Fatalf("Should fail as xz is unsupported for compression format.")
+	if _, err := CompressStream(&bytes.Buffer{}, Bzip2); err == nil {
+		t.Fatalf("Should fail as bz2 is unsupported for compression format.")
 	}
 }
 
 func TestCompressStreamInvalid(t *testing.T) {
-	dest, err := os.Create(tmp + "dest")
-	if err != nil {
-		t.Fatalf("Fail to create the destination file: %v", err)
-	}
-	defer dest.Close()
-
-	_, err = CompressStream(dest, -1)
-	if err == nil {
-		t.Fatalf("Should fail as xz is unsupported for compression format.")
+	if _, err := CompressStream(&bytes.Buffer{}, -1); err == nil {
+		t.Fatalf("Should fail as -1 is an invalid compression format.")
 	}
 }
 
@@ -213,10 +193,7 @@ func TestExtensionXz(t *testing.T) {
 
 func createEmptyFile(t *testing.T, path string) {
 	t.Helper()
-	f, err := os.Create(path)
-	require.NoError(t, err)
-	err = f.Close()
-	require.NoError(t, err)
+	require.NoError(t, os.WriteFile(path, nil, 0o666))
 }
 
 func TestUntarPathWithInvalidDest(t *testing.T) {
@@ -771,7 +748,7 @@ func TestTypeXGlobalHeaderDoesNotFail(t *testing.T) {
 	hdr := tar.Header{Typeflag: tar.TypeXGlobalHeader}
 	tmpDir := t.TempDir()
 	buffer := make([]byte, 1<<20)
-	err := createTarFile(filepath.Join(tmpDir, "pax_global_header"), tmpDir, &hdr, nil, true, nil, false, false, nil, buffer)
+	err := extractTarFileEntry(filepath.Join(tmpDir, "pax_global_header"), tmpDir, &hdr, nil, true, nil, false, false, nil, buffer)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -810,7 +787,7 @@ func TestUntarUstarGnuConflict(t *testing.T) {
 
 func prepareUntarSourceDirectory(numberOfFiles int, targetPath string, makeLinks bool) (int, error) {
 	fileData := []byte("fooo")
-	for n := 0; n < numberOfFiles; n++ {
+	for n := range numberOfFiles {
 		fileName := fmt.Sprintf("file-%d", n)
 		if err := os.WriteFile(filepath.Join(targetPath, fileName), fileData, 0o700); err != nil {
 			return 0, err
@@ -1137,7 +1114,7 @@ func TestTempArchiveCloseMultipleTimes(t *testing.T) {
 	if n != 5 {
 		t.Fatalf("Expected to read 5 bytes. Read %d instead", n)
 	}
-	for i := 0; i < 3; i++ {
+	for i := range 3 {
 		if err = tempArchive.Close(); err != nil {
 			t.Fatalf("i=%d. Unexpected error closing temp archive: %v", i, err)
 		}
@@ -1247,3 +1224,88 @@ func readFileFromArchive(t *testing.T, archive io.ReadCloser, name string, expec
 	assert.NoError(t, err)
 	return string(content)
 }
+
+func TestTimestamp(t *testing.T) {
+	// write single file into dir that we'll tar
+	td := t.TempDir()
+	tf := filepath.Join(td, "foo")
+
+	require.NoError(t, os.WriteFile(tf, []byte("bar"), 0o644))
+
+	// helper function to tar that dir and return byte slice
+	tarToByteSlice := func(options *TarOptions) []byte {
+		rc, err := TarWithOptions(td, options)
+		assert.NoError(t, err)
+		defer rc.Close()
+
+		rv, err := io.ReadAll(rc)
+		assert.NoError(t, err)
+		return rv
+	}
+
+	// default options
+	defaultOptions := &TarOptions{}
+
+	// override timestamp option
+	epochOptions := &TarOptions{Timestamp: &time.Time{}}
+
+	// get tar bytes slices now
+	origTarDefaultOptions := tarToByteSlice(defaultOptions)
+	origTarEpochOptions := tarToByteSlice(epochOptions)
+
+	// set the mod time of the file to an hour later
+	oneHourLater := time.Now().Add(time.Hour)
+	require.NoError(t, os.Chtimes(tf, oneHourLater, oneHourLater))
+
+	// then tar again
+	laterTarDefaultOptions := tarToByteSlice(defaultOptions)
+	laterTarEpochOptions := tarToByteSlice(epochOptions)
+
+	// we expect the ones without a fixed timestamp to be different
+	assert.NotEqual(t, origTarDefaultOptions, laterTarDefaultOptions)
+
+	// we expect the ones with a fixed timestamp to be the same
+	assert.Equal(t, origTarEpochOptions, laterTarEpochOptions)
+}
+
+type errorBuf struct {
+	bytes.Buffer
+
+	err        error
+	failWrite  int
+	writeCount int
+}
+
+func (b *errorBuf) Write(d []byte) (int, error) {
+	b.writeCount++
+	if b.failWrite == b.writeCount {
+		return 0, b.err
+	}
+	return b.Buffer.Write(d)
+}
+
+func (b *errorBuf) Close() error {
+	return nil
+}
+
+func TestTarErrorHandling(t *testing.T) {
+	dir := t.TempDir()
+
+	for i := range 10 {
+		if err := os.WriteFile(filepath.Join(dir, fmt.Sprintf("file%d", i)), bytes.Repeat([]byte("hello"), 32<<10),
+			0o700); err != nil {
+			t.Fatal(err)
+		}
+	}
+
+	dest := &errorBuf{
+		failWrite: 1,
+		err:       errors.New("boom"),
+	}
+
+	if err := tarWithOptionsTo(dest, dir, &TarOptions{
+		Compression: Uncompressed,
+	}); !errors.Is(err, dest.err) {
+		t.Fatalf("Did not propagate error; got %v", err)
+	}
+}
